# -*- coding: UTF-8 -*-
###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# libsecure.py
#
# classes utilitaires pour lancement des services en https
# et vérification des certificats
#
###########################################################################

import sys, os
from glob import glob
import httplib, xmlrpclib
from twisted.python import log

# SSL imports
from OpenSSL import SSL
import sys, xmlrpclib
import traceback
from time import time
from urlparse import urlparse

try:
    # Get C binding version
    from OpenSSL.crypto import X509_verify_cert_error_string
except ImportError:
    try:
        # ctypes binding version
        from ctypes import cdll, c_ulong, c_long, c_char_p, c_int, create_string_buffer
        from ctypes.util import find_library

        _libssl = cdll.LoadLibrary(find_library('ssl'))

        X509_verify_cert_error_string = _libssl.X509_verify_cert_error_string
        X509_verify_cert_error_string.restype = c_char_p
        X509_verify_cert_error_string.argtypes = (c_long,)

    except (ImportError, AttributeError, TypeError):
        # ctypes module or X509_verify_cert_error_string symbol is missing:
        # use dummy fallback
        def X509_verify_cert_error_string(errnum):
            return str(errnum)

# transport sécurisé utilisant un certificat
class TransportEole(xmlrpclib.SafeTransport):

    def __init__(self, cert_file, key_file = None):
        self.cert_file = cert_file
        self.key_file = key_file or cert_file
        super(xmlrpclib.SafeTransport, self).__init__()

    def make_connection(self, host):
        # create a HTTPS connection object from a host descriptor
        # host may be a string, or a (host, x509-dict) tuple
        localhost, extra_headers, x509 = self.get_host_info(host)
        try:
            HTTPS = httplib.HTTPS
        except AttributeError:
            raise NotImplementedError(
                "your version of httplib doesn't support HTTPS"
                )
        else:
            cx = HTTPS(localhost, None,
                       key_file = self.cert_file,
                       cert_file = self.cert_file)
            return cx

class ClientContextFactory:
    """context factory for SSL clients."""
    isClient = 1
    method = SSL.SSLv3_METHOD

    def __init__(self, certfile, keyfile='', ca_location='', mode=None):
        self.certfile = certfile
        self.keyfile = keyfile
        self.ca_location = ca_location
        if mode is None:
            self.verify_mode = SSL.VERIFY_PEER|SSL.VERIFY_FAIL_IF_NO_PEER_CERT
        else:
            self.verify_mode = mode
        self.peer_cert_data = {}

    # fonction de vérification d'un certificat (utilisé pour les proxies)
    def _verify(self, connection, x509, errnum, errdepth, ok):
        """ This function will be called once for each certificate in the chain
        """
        self.peer_cert_data['subject'] = x509.get_subject()
        self.peer_cert_data['issuer'] = x509.get_issuer()
        self.peer_cert_data['expired'] = x509.has_expired()
        self.peer_cert_data['error'] = errnum
        if errnum != 0:
            # error code 0 -> OK
            log.msg(_('Proxy certificate validation : Error %s at depth %s : %s') % (errnum, errdepth, X509_verify_cert_error_string(errnum)))
        return ok

    def getContext(self):
        ctx = SSL.Context(self.method)
        ctx.use_certificate_file(self.certfile)
        try:
            if self.keyfile and os.path.isfile(self.keyfile):
                ctx.use_privatekey_file(self.keyfile)
            else:
                ctx.use_privatekey_file(self.certfile)
        except:
            # impossible de charger la clé depuis le fichier du certificat, on regarde si il n'y a pas un fichier .key associé
            alt_keyfile = os.path.splitext(self.certfile)[0] + '.key'
            ctx.use_privatekey_file(alt_keyfile)
        # chargement des certificats des autorités de certification (si disponible)
        if os.path.isdir(self.ca_location):
            ca_certs = [os.path.join(self.ca_location, ca_file) for ca_file in os.listdir(self.ca_location)]
        elif os.path.isfile(self.ca_location):
            ca_certs = [self.ca_location]
        else:
            log.msg(_("Invalid CA location"), self.ca_location)
        for ca_cert in ca_certs:
            if os.path.splitext(ca_cert)[1] in [".pem",".der",".crt",".cert"]:
                try:
                    ctx.load_verify_locations(ca_cert)
                except SSL.Error, e:
                    log.msg(_('Error loading CA certificate %s : %s') % (ca_cert, str(e)))
        ctx.set_verify(self.verify_mode, self._verify)
        ctx.set_verify_depth(10)
        return ctx

class ServerContextFactory:
    """
    Factory permettant de créer un contexte SSL
    """

    def __init__(self, certfile, keyfile='', ca_file=''):
        self.certfile = certfile
        self.keyfile = keyfile
        self.ca_file = ca_file

    def getContext(self):
        """Create an SSL context.
        This is a sample implementation that loads a certificate from a file
        called 'server.pem'."""
        ctx = SSL.Context(SSL.SSLv23_METHOD)
        ctx.use_certificate_file(self.certfile)
        try:
            if self.keyfile:
                ctx.use_privatekey_file(self.keyfile)
            else:
                ctx.use_privatekey_file(self.certfile)
        except:
            # impossible de charger la clé depuis le fichier du certificat, on regarde si il n'y a pas un fichier .key associé
            alt_keyfile = os.path.splitext(self.certfile)[0] + '.key'
            try:
                ctx.use_privatekey_file(alt_keyfile)
            except SSL.Error, e:
                log.msg(_('Error loading Key file %s : %s') % (alt_keyfile, str(e)))
        if self.ca_file:
            try:
                # chargement des ca à envoyer au client
                ctx.load_client_ca(self.ca_file)
            except Exception, e:
                traceback.print_exc()
                log.msg(_('Error loading certificates chain file %s : %s') % (self.ca_file, str(e)))
        return ctx

